#include"net_server.h"

#include"enet.h"
#include<stdio.h>
#include"luaapi.h"
#include"net.h"
#include"game.h"
#include"resman.h"
#include<cglm/vec3.h>

static ENetHost *host;
static ENetPeer *peer;

static uint16_t serverTick;

#define RECONBUFSZ 64
static size_t stateUpdateLastClientTick;
static ENetPacket *stateUpdates[RECONBUFSZ];
static struct CPlayerCtrl myInputs[RECONBUFSZ];
static int dejitterLeft = 0;

void net_client_init() {
	host = enet_host_create(NULL, 1, 1, 0, 0);
	
	Game.isMultiplayer = 1;
	Game.isAuthority = 0;
}

bool net_client_connect(ENetAddress eaddr) {
	peer = enet_host_connect(host, &eaddr, 1, 0);
	
	ENetEvent ev;
	if(enet_host_service(host, &ev, 200) <= 0 || ev.type != ENET_EVENT_TYPE_CONNECT) {
		enet_peer_reset(peer);
		return false;
	}
	
	return true;
}

extern double glfwGetTime();

static void interpret_pkt(ENetPacket *pkt) {
	struct bstr b;
	bwrap(&b, pkt->data, pkt->dataLength);
	
	uint16_t cmdCount = b_ru16(&b) & ~PKT_PERFRAME;
	uint16_t t = b_ru16(&b);
	
	for(size_t i = 0; i < cmdCount; i++) {
		uint16_t cmd = b_ru16(&b);
		if(cmd == CMD_SC_ASSIGN) {
			Game.spectated = Game.controlled = b_ru16(&b);
		} else if(cmd == CMD_SC_SINGLE_COMPONENT) {
			uint8_t ctype = b_ru8(&b);
			
			uint16_t ent = b_ru16(&b);
			
			if(ctype == CMD_CTYPE_RENDER) {
				struct CRender *c = game_ensurecomponent(ent, render);
				const char *mdl = b_rzstr(&b);
				if(strncmp(c->mdl, mdl, MODELNAME_LENGTH)) {
					if(strlen(c->mdl)) {
						resman_unref(RESMAN_MODEL, c->mdl);
						c->cache = NULL;
					}
					memcpy(c->mdl, mdl, MODELNAME_LENGTH);
				}
			} else if(ctype == CMD_CTYPE_PHYSICS) {
				struct CPhysics *c = game_ensurecomponent(ent, physics);
				c->trigger = b_ru16(&b);
				c->type = b_ru8(&b);
				c->dynamics = b_ru8(&b);
				if(c->type == CPHYSICS_BOX) {
					c->box.w = b_rf32(&b);
					c->box.h = b_rf32(&b);
					c->box.l = b_rf32(&b);
				} else if(c->type == CPHYSICS_CAPSULE) {
					c->capsule.length = b_rf32(&b);
					c->capsule.radius = b_rf32(&b);
				} else abort();
				
				c->start[0] = b_rf32(&b);
				c->start[1] = b_rf32(&b);
				c->start[2] = b_rf32(&b);
				
				if((c->dynamics & ~CPHYSICS_GHOST) == CPHYSICS_DYNAMIC) {
					c->vel[0] = b_rf32(&b);
					c->vel[1] = b_rf32(&b);
					c->vel[2] = b_rf32(&b);
					
					c->speed = b_rf32(&b);
				}
				
				c->mass = b_rf32(&b);
				c->friction = b_rf32(&b);
				c->collide = b_ru32(&b);
				
				if(c->geom) {
					dGeomSetPosition(c->geom, c->start[0], c->start[1], c->start[2]);
					
					if((c->dynamics & ~CPHYSICS_GHOST) == CPHYSICS_DYNAMIC) {
						dBodySetLinearVel(dGeomGetBody(c->geom), c->vel[0], c->vel[1], c->vel[2]);
					}
				}
			} else if(ctype == CMD_CTYPE_MOVEMENT) {
				struct CMovement *c = game_ensurecomponent(ent, movement);
				c->dir[0] = b_rf32(&b);
				c->dir[1] = b_rf32(&b);
				c->dir[2] = b_rf32(&b);
				c->pointing = b_rf32(&b);
				c->jump = b_ru8(&b);
				c->canJump = b_ru8(&b);
				c->holding = b_ru16(&b);
			} if(ctype == CMD_CTYPE_BONED) {
				struct CBoned *c = game_ensurecomponent(ent, boned);
				c->anim.standard = b_ru8(&b);
			}
		} else if(cmd == CMD_SC_POSUPDATE) {
			uint16_t lastOurTick = b_ru16(&b);
			
			stateUpdateLastClientTick = stateUpdateLastClientTick < lastOurTick ? lastOurTick : stateUpdateLastClientTick + 1;
			
			uint16_t conveyorCount = b_ru16(&b);
			if(conveyorCount > Game.conveyorCount) conveyorCount = Game.conveyorCount;
			for(size_t i = 0; i < conveyorCount; i++) {
				struct Conveyor *conveyor = Game.conveyors[i];
				conveyor->position = b_rf16(&b);
			}
			
			for(size_t i = 0; i < Game.entities.physicsCount; i++) {
				vec3 p = {b_rf16(&b), b_rf16(&b), b_rf16(&b)};
				if(Game.entities.physics[i].geom) {
					dGeomSetPosition(Game.entities.physics[i].geom, p[0], p[1], p[2]);
				} else {
					glm_vec3_copy(p, Game.entities.physics[i].start);
				}
				
				if((Game.entities.physics[i].dynamics & ~CPHYSICS_GHOST) == CPHYSICS_DYNAMIC) {
					vec3 v = {b_rf16(&b), b_rf16(&b), b_rf16(&b)};
					
					if(Game.entities.physics[i].geom) {
						dBodyID bid = dGeomGetBody(Game.entities.physics[i].geom);
						dBodySetLinearVel(bid, v[0], v[1], v[2]);
					} else {
						glm_vec3_copy(v, Game.entities.physics[i].vel);
					}
					
					struct CMovement *cm = game_getcomponent(Game.entities.physics[i].entity, movement);
					if(cm) {
						glm_vec3_normalize_to(v, cm->dir);
					}
					
					if(Game.entities.physics[i].geom) {
						vec4 r = {0, b_rf16(&b), b_rf16(&b), b_rf16(&b)};
						r[0] = b_rf16(&b);
						
						dGeomSetQuaternion(Game.entities.physics[i].geom, (float*) r);
					} else {
						Game.entities.physics[i].rot[0] = b_rf16(&b);
						Game.entities.physics[i].rot[1] = b_rf16(&b);
						Game.entities.physics[i].rot[2] = b_rf16(&b);
						Game.entities.physics[i].rot[3] = b_rf16(&b);
					}
				}
			}
			
			for(size_t i = 0; i < Game.entities.movementCount; i++) {
				Game.entities.movement[i].pointing = b_rf16(&b);
			}
			
			if(Game.reconciliate) {
	//			printf("Recon %lu -> %lu\n", stateUpdateLastClientTick, Game.frame);
				if(Game.frame - stateUpdateLastClientTick < RECONBUFSZ) {
					size_t properFrame = Game.frame;
					
					struct CPlayerCtrl *cc = &myInputs[Game.frame - stateUpdateLastClientTick];
					struct CPlayerCtrl *w = game_getcomponent(Game.controlled, playerctrl);
					size_t keys = w ? w->keys : 0;
	//				printf("Currently %X\n", keys);
					for(; cc > myInputs + 1; cc--) {
						w = game_getcomponent(Game.controlled, playerctrl);
						if(w) w->keys = cc->keys;
	//				printf("pred      %X\n", cc->keys);
						
						game_update();
					}
					
					w = game_getcomponent(Game.controlled, playerctrl);
					if(w) w->keys = keys;
					
					Game.frame = properFrame;
				}
			}
		} else if(cmd == CMD_SC_LOAD_SCRIPT) {
			luaapi_load(b_rzstr(&b));
		} else if(cmd == CMD_SC_MSG) {
			if(!luaapi_recvmsg(&b, -2)) {
				// Corrupted packet.
				break;
			}
		}
	}
}

void net_client_receive() {
	ENetEvent ev;
	while(enet_host_service(host, &ev, 0) > 0) {
		if(ev.type == ENET_EVENT_TYPE_CONNECT) {
			luaapi_cleanup();
		} else if(ev.type == ENET_EVENT_TYPE_RECEIVE) {
			struct bstr b;
			bwrap(&b, ev.packet->data, ev.packet->dataLength);
			
			uint16_t cmdCount = b_ru16(&b);
			int perframe = cmdCount & PKT_PERFRAME;
			cmdCount &= ~PKT_PERFRAME;
			
			uint16_t t = b_ru16(&b);
			
			if(perframe) {
				if(stateUpdates[RECONBUFSZ - 1]) {
					enet_packet_destroy(stateUpdates[RECONBUFSZ - 1]);
				}
				memmove(stateUpdates + 1, stateUpdates, sizeof(*stateUpdates) * (RECONBUFSZ - 1));
				stateUpdates[0] = ev.packet;
				
				dejitterLeft++;
				if(dejitterLeft > RECONBUFSZ) dejitterLeft = RECONBUFSZ;
			} else {
				interpret_pkt(ev.packet);
				enet_packet_destroy(ev.packet);
			}
		}
	}
}

void net_client_dejitter() {
	if(dejitterLeft) {
		interpret_pkt(stateUpdates[--dejitterLeft]);
	}
}

void net_client_update() {
	struct CPlayerCtrl *cc = game_getcomponent(Game.controlled, playerctrl);
	
	memmove(myInputs + 1, myInputs, sizeof(*myInputs) * (RECONBUFSZ - 1));
	
	if(cc) {
		memcpy(myInputs, cc, sizeof(*cc));
		
		struct bstr b;
		bstr(&b, 64);
		
		b_wu16(&b, 1);
		b_wu16(&b, Game.frame);
		
		b_wu16(&b, CMD_SC_SINGLE_COMPONENT);
		b_wu8(&b, CMD_CTYPE_PLAYERCTRL);
		
		b_wu16(&b, Game.controlled);
		
		b_wf32(&b, cc->yaw);
		b_wf32(&b, cc->pitch);
		b_wu32(&b, cc->keys);
		
		enet_peer_send(peer, 0, enet_packet_create(b.data, b.wi, ENET_PACKET_FLAG_RELIABLE));
		free(b.data);
		
		enet_host_flush(host);
	} else {
		memset(myInputs, 0, sizeof(*cc));
	}
}

void net_client_send_msg(struct bstr *data) {
	struct bstr b;
	bstr(&b, 64);
	
	b_wu16(&b, 1);
	b_wu16(&b, Game.frame);
	
	b_wu16(&b, CMD_SC_MSG);
	b_wraw(&b, data->data, data->wi);
	
	enet_peer_send(peer, 0, enet_packet_create(b.data, b.wi, ENET_PACKET_FLAG_RELIABLE));
	
	free(b.data);
	
	enet_host_flush(host);
}
